@@ -4,6 +4,7 @@ Allows for selecting data subsets using certain latitude and longitude
 bounds; selecting geographical regions; constructing area averages; etc.
 """
 import logging
+import warnings
 
 import fiona
 import iris
@@ -13,6 +14,11 @@ import shapely.ops
 from dask import array as da
 from iris.exceptions import CoordinateNotFoundError
 
+from ._ancillary_vars import (
+    add_ancillary_variable,
+    add_cell_measure,
+    remove_fx_variables,
+)
 from ._shared import (
     get_iris_analysis_operation,
     guess_bounds,
@@ -21,6 +27,8 @@ from ._shared import (
 
 logger = logging.getLogger(__name__)
 
+SHAPE_ID_KEYS = ('name', 'NAME', 'Name', 'id', 'ID')
+
 
 def extract_region(cube, start_longitude, end_longitude, start_latitude,
                    end_latitude):
@@ -177,6 +185,23 @@ def meridional_statistics(cube, operator):
     raise ValueError(msg)
 
 
+def compute_area_weights(cube):
+    """Compute area weights."""
+    with warnings.catch_warnings(record=True) as caught_warnings:
+        warnings.filterwarnings(
+            'always',
+            message="Using DEFAULT_SPHERICAL_EARTH_RADIUS.",
+            category=UserWarning,
+            module='iris.analysis.cartography',
+        )
+        weights = iris.analysis.cartography.area_weights(cube)
+        for warning in caught_warnings:
+            logger.debug(
+                "%s while computing area weights of the following cube:\n%s",
+                warning.message, cube)
+    return weights
+
+
 def area_statistics(cube, operator):
     """Apply a statistical operator in the horizontal direction.
 
@@ -232,11 +257,12 @@ def area_statistics(cube, operator):
     try:
         grid_areas = cube.cell_measure('cell_area').core_data()
     except iris.exceptions.CellMeasureNotFoundError:
-        logger.info(
+        logger.debug(
             'Cell measure "cell_area" not found in cube %s. '
-            'Check fx_file availability.', cube.summary(shorten=True)
-        )
-        logger.info('Attempting to calculate grid cell area...')
+            'Check fx_file availability.', cube.summary(shorten=True))
+        logger.debug('Attempting to calculate grid cell area...')
+    else:
+        grid_areas = da.broadcast_to(grid_areas, cube.shape)
 
     if grid_areas is None and cube.coord('latitude').points.ndim == 2:
         coord_names = [coord.standard_name for coord in cube.coords()]
@@ -247,8 +273,8 @@ def area_statistics(cube, operator):
             cube_tmp.coord('grid_latitude').rename('latitude')
             cube_tmp.remove_coord('longitude')
             cube_tmp.coord('grid_longitude').rename('longitude')
-            grid_areas = iris.analysis.cartography.area_weights(cube_tmp)
-            logger.info('Calculated grid area shape: %s', grid_areas.shape)
+            grid_areas = compute_area_weights(cube_tmp)
+            logger.debug('Calculated grid area shape: %s', grid_areas.shape)
         else:
             logger.error(
                 'fx_file needed to calculate grid cell area for irregular '
@@ -259,8 +285,8 @@ def area_statistics(cube, operator):
     coord_names = ['longitude', 'latitude']
     if grid_areas is None:
         cube = guess_bounds(cube, coord_names)
-        grid_areas = iris.analysis.cartography.area_weights(cube)
-        logger.info('Calculated grid area shape: %s', grid_areas.shape)
+        grid_areas = compute_area_weights(cube)
+        logger.debug('Calculated grid area shape: %s', grid_areas.shape)
 
     if cube.shape != grid_areas.shape:
         raise ValueError('Cube shape ({}) doesn`t match grid area shape '
@@ -279,7 +305,7 @@ def area_statistics(cube, operator):
 
     new_dtype = result.dtype
     if original_dtype != new_dtype:
-        logger.warning(
+        logger.debug(
             "area_statistics changed dtype from "
             "%s to %s, changing back", original_dtype, new_dtype)
         result.data = result.core_data().astype(original_dtype)
@@ -405,8 +431,12 @@ def _correct_coords_from_shapefile(cube, cmor_coords, pad_north_pole,
     return lon, lat
 
 
-def _get_masks_from_geometries(geometries, lon, lat, method='contains',
-                               decomposed=False, ids=None):
+def _get_masks_from_geometries(geometries,
+                               lon,
+                               lat,
+                               method='contains',
+                               decomposed=False,
+                               ids=None):
 
     if method not in {'contains', 'representative'}:
         raise ValueError(
@@ -417,7 +447,7 @@ def _get_masks_from_geometries(geometries, lon, lat, method='contains',
     if ids:
         ids = [str(id_) for id_ in ids]
     for i, item in enumerate(geometries):
-        for id_prop in ('name', 'NAME', 'Name', 'id', 'ID'):
+        for id_prop in SHAPE_ID_KEYS:
             if id_prop in item['properties']:
                 id_ = str(item['properties'][id_prop])
                 break
@@ -439,6 +469,53 @@ def _get_masks_from_geometries(geometries, lon, lat, method='contains',
     return selections
 
 
+def _geometry_matches_ids(geometry: dict, ids: list):
+    """Returns True if `geometry` matches one of the `ids`."""
+    props = geometry['properties']
+
+    geom_id = [props.get(key, None) for key in SHAPE_ID_KEYS]
+    geom_id = [key for key in geom_id if key is not None]
+
+    if not geom_id:
+        raise KeyError(f'{props} dict has no `name` or `id` key')
+
+    geom_id = geom_id[0]
+
+    return geom_id in ids
+
+
+def _get_bounds(geometries, ids=None):
+    """Get bounds from the subset of geometries defined by `ids`.
+
+    Parameters
+    ----------
+    geometries : fiona.Collection
+        Fiona collection of shapes (geometries).
+    ids : tuple of str, optional
+        List of ids to select from geometry collection. If None,
+        return global bounds (``geometries.bounds``)
+
+    Returns
+    -------
+    lat_min, lon_min, lat_max, lon_max
+        Returns coordinates deliminating bounding box for shape ids.
+    """
+    if not ids:
+        return geometries.bounds
+
+    subset = [geom for geom in geometries if _geometry_matches_ids(geom, ids)]
+
+    all_points = [
+        np.hstack(geom['geometry']['coordinates']) for geom in subset
+    ]
+    all_points = np.vstack(all_points)
+
+    lon_max, lat_max = all_points.max(axis=0)
+    lon_min, lat_min = all_points.min(axis=0)
+
+    return lon_min, lat_min, lon_max, lat_max
+
+
 def _get_shape(lon, lat, method, item):
     shape = shapely.geometry.shape(item['geometry'])
     if method == 'contains':
@@ -496,8 +573,12 @@ def fix_coordinate_ordering(cube):
     return cube
 
 
-def extract_shape(cube, shapefile, method='contains', crop=True,
-                  decomposed=False, ids=None):
+def extract_shape(cube,
+                  shapefile,
+                  method='contains',
+                  crop=True,
+                  decomposed=False,
+                  ids=None):
     """Extract a region defined by a shapefile.
 
     Note that this function does not work for shapes crossing the
@@ -551,8 +632,15 @@ def extract_shape(cube, shapefile, method='contains', crop=True,
             pad_hawaii = True
 
         if crop:
+            lon_min, lat_min, lon_max, lat_max = _get_bounds(
+                geometries=geometries,
+                ids=ids,
+            )
             cube = _crop_cube(cube,
-                              *geometries.bounds,
+                              start_longitude=lon_min,
+                              start_latitude=lat_min,
+                              end_longitude=lon_max,
+                              end_latitude=lat_max,
                               cmor_coords=cmor_coords)
 
         lon, lat = _correct_coords_from_shapefile(cube, cmor_coords,
@@ -572,9 +660,17 @@ def _mask_cube(cube, selections):
     cubelist = iris.cube.CubeList()
     for id_, select in selections.items():
         _cube = cube.copy()
+        remove_fx_variables(_cube)
         _cube.add_aux_coord(
             iris.coords.AuxCoord(id_, units='no_unit', long_name="shape_id"))
         select = da.broadcast_to(select, _cube.shape)
         _cube.data = da.ma.masked_where(~select, _cube.core_data())
         cubelist.append(_cube)
-    return fix_coordinate_ordering(cubelist.merge_cube())
+    result = fix_coordinate_ordering(cubelist.merge_cube())
+    if cube.cell_measures():
+        for measure in cube.cell_measures():
+            add_cell_measure(result, measure, measure.measure)
+    if cube.ancillary_variables():
+        for ancillary_variable in cube.ancillary_variables():
+            add_ancillary_variable(result, ancillary_variable)
+    return result
