@@ -1,5 +1,12 @@
-"""Functions to compute multi-cube statistics."""
+"""Statistics across cubes.
 
+This module contains functions to compute statistics across multiple cubes or
+products.
+
+Wrapper functions separate esmvalcore internals, operating on products, from
+generalized functions that operate on iris cubes. These wrappers support
+grouped execution by passing a groupby keyword.
+"""
 import logging
 import re
 import warnings
@@ -8,10 +15,15 @@ from functools import reduce
 
 import cf_units
 import iris
+import iris.coord_categorisation
 import numpy as np
 from iris.util import equalise_attributes
+
+from esmvalcore.iris_helpers import date2num
 from esmvalcore.preprocessor import remove_fx_variables
 
+from ._other import _group_products
+
 logger = logging.getLogger(__name__)
 
 STATISTIC_MAPPING = {
@@ -78,13 +90,13 @@ def _get_consistent_time_unit(cubes):
 def _unify_time_coordinates(cubes):
     """Make sure all cubes' share the same time coordinate.
 
-    This function extracts the date information from the cube and
-    reconstructs the time coordinate, resetting the actual dates to the
-    15th of the month or 1st of july for yearly data (consistent with
-    `regrid_time`), so that there are no mismatches in the time arrays.
+    This function extracts the date information from the cube and reconstructs
+    the time coordinate, resetting the actual dates to the 15th of the month or
+    1st of July for yearly data (consistent with `regrid_time`), so that there
+    are no mismatches in the time arrays.
 
-    If cubes have different time units, it will use reset the calendar to
-    a default gregorian calendar with unit "days since 1850-01-01".
+    If cubes have different time units, it will reset the calendar to a
+    default gregorian calendar with unit "days since 1850-01-01".
 
     Might not work for (sub)daily data, because different calendars may have
     different number of days in the year.
@@ -125,7 +137,7 @@ def _unify_time_coordinates(cubes):
                 "support sub-daily data.")
 
         # Update the cubes' time coordinate (both point values and the units!)
-        cube.coord('time').points = t_unit.date2num(dates)
+        cube.coord('time').points = date2num(dates, t_unit, coord.dtype)
         cube.coord('time').units = t_unit
         cube.coord('time').bounds = None
         cube.coord('time').guess_bounds()
@@ -153,13 +165,33 @@ def _map_to_new_time(cube, time_points):
     sample_points = [('time', time_points)]
     scheme = iris.analysis.Nearest(extrapolation_mode='mask')
 
+    # Make sure that all integer time coordinates ('year', 'month',
+    # 'day_of_year', etc.) are converted to floats, otherwise the
+    # nearest-neighbor interpolation will fail with a "cannot convert float NaN
+    # to integer". In addition, remove their bounds (this would be done by iris
+    # anyway).
+    int_time_coords = []
+    for coord in cube.coords(dimensions=cube.coord_dims('time'),
+                             dim_coords=False):
+        if np.issubdtype(coord.points.dtype, np.integer):
+            int_time_coords.append(coord.name())
+            coord.points = coord.points.astype(float)
+            coord.bounds = None
+
+    # Do the actual interpolation
     try:
         new_cube = cube.interpolate(sample_points, scheme)
     except Exception as excinfo:
         raise ValueError(
-            "Tried to align cubes in multi-model statistics, but failed for"
-            f" cube {cube} and time points {time_points}. Encountered the "
-            f"following exception: {excinfo}")
+            f"Tried to align cubes in multi-model statistics, but failed for "
+            f"cube {cube}\n and time points {time_points}") from excinfo
+
+    # Change the dtype of int_time_coords to their original values
+    for coord_name in int_time_coords:
+        coord = new_cube.coord(coord_name,
+                               dimensions=new_cube.coord_dims('time'),
+                               dim_coords=False)
+        coord.points = coord.points.astype(int)
 
     return new_cube
 
@@ -191,29 +223,71 @@ def _align(cubes, span):
     return new_cubes
 
 
+def _equalise_cell_methods(cubes):
+    """Equalise cell methods in cubes (in-place)."""
+    # Simply remove all cell methods
+    for cube in cubes:
+        cube.cell_methods = None
+
+
+def _equalise_coordinates(cubes):
+    """Equalise coordinates in cubes (in-place)."""
+    if not cubes:
+        return
+
+    # If metadata of a coordinate metadata is equal for all cubes, do not
+    # modify it; else remove long_name and attributes.
+    equal_coords_metadata = []
+    for coord in cubes[0].coords():
+        for other_cube in cubes[1:]:
+            other_cube_has_equal_coord = [
+                coord.metadata == other_coord.metadata for other_coord in
+                other_cube.coords(coord.name())
+            ]
+            if not any(other_cube_has_equal_coord):
+                break
+        else:
+            equal_coords_metadata.append(coord.metadata)
+
+    # Modify coordinates accordingly
+    for cube in cubes:
+        for coord in cube.coords():
+            if coord.metadata not in equal_coords_metadata:
+                coord.long_name = None
+                coord.attributes = None
+
+        # Additionally remove specific scalar coordinates which are not
+        # expected to be equal in the input cubes
+        scalar_coords_to_remove = ['p0', 'ptop']
+        for scalar_coord in cube.coords(dimensions=()):
+            if scalar_coord.var_name in scalar_coords_to_remove:
+                cube.remove_coord(scalar_coord)
+
+
+def _equalise_fx_variables(cubes):
+    """Equalise fx variables in cubes (in-place)."""
+    # Simple remove all fx variables
+    for cube in cubes:
+        remove_fx_variables(cube)
+
+
 def _combine(cubes):
     """Merge iris cubes into a single big cube with new dimension.
 
     This assumes that all input cubes have the same shape.
     """
-    equalise_attributes(cubes)  # in-place
+    # Equalise some metadata that can cause merge to fail (in-place)
+    # https://scitools-iris.readthedocs.io/en/stable/userguide/
+    #    merge_and_concat.html#common-issues-with-merge-and-concatenate
+    equalise_attributes(cubes)
+    _equalise_cell_methods(cubes)
+    _equalise_coordinates(cubes)
+    _equalise_fx_variables(cubes)
 
     for i, cube in enumerate(cubes):
         concat_dim = iris.coords.AuxCoord(i, var_name=CONCAT_DIM)
-
         cube.add_aux_coord(concat_dim)
 
-        # Clear some metadata that can cause merge to fail
-        # https://scitools-iris.readthedocs.io/en/stable/userguide/
-        #    merge_and_concat.html#common-issues-with-merge-and-concatenate
-
-        remove_fx_variables(cube)
-        cube.cell_methods = None
-
-        for coord in cube.coords():
-            coord.long_name = None
-            coord.attributes = None
-
     cubes = iris.cube.CubeList(cubes)
 
     merged_cube = cubes.merge_cube()
@@ -221,14 +295,32 @@ def _combine(cubes):
     return merged_cube
 
 
+def _compute_slices(cubes):
+    """Create cube slices resulting in a combined cube of about 1 GiB."""
+    gibibyte = 2**30
+    total_bytes = cubes[0].data.nbytes * len(cubes)
+    n_slices = int(np.ceil(total_bytes / gibibyte))
+
+    n_timesteps = cubes[0].shape[0]
+    slice_len = int(np.ceil(n_timesteps / n_slices))
+
+    for i in range(n_slices):
+        start = i * slice_len
+        end = (i + 1) * slice_len
+        if end >= n_timesteps:
+            yield slice(start, n_timesteps)
+            return
+        yield slice(start, end)
+
+
 def _compute_eager(cubes: list, *, operator: iris.analysis.Aggregator,
                    **kwargs):
     """Compute statistics one slice at a time."""
     _ = [cube.data for cube in cubes]  # make sure the cubes' data are realized
 
     result_slices = []
-    for i in range(cubes[0].shape[0]):
-        single_model_slices = [cube[i] for cube in cubes]
+    for chunk in _compute_slices(cubes):
+        single_model_slices = [cube[chunk] for cube in cubes]
         combined_slice = _combine(single_model_slices)
         with warnings.catch_warnings():
             warnings.filterwarnings(
@@ -250,7 +342,7 @@ def _compute_eager(cubes: list, *, operator: iris.analysis.Aggregator,
         result_slices.append(collapsed_slice)
 
     try:
-        result_cube = iris.cube.CubeList(result_slices).merge_cube()
+        result_cube = iris.cube.CubeList(result_slices).concatenate_cube()
     except Exception as excinfo:
         raise ValueError(
             "Multi-model statistics failed to concatenate results into a"
@@ -261,6 +353,15 @@ def _compute_eager(cubes: list, *, operator: iris.analysis.Aggregator,
 
     result_cube.data = np.ma.array(result_cube.data)
     result_cube.remove_coord(CONCAT_DIM)
+    if result_cube.cell_methods:
+        cell_method = result_cube.cell_methods[0]
+        result_cube.cell_methods = None
+        updated_method = iris.coords.CellMethod(
+            method=cell_method.method,
+            coords=cell_method.coord_names,
+            intervals=cell_method.intervals,
+            comments=f'input_cubes: {len(cubes)}')
+        result_cube.add_cell_method(updated_method)
     return result_cube
 
 
@@ -287,7 +388,6 @@ def _multicube_statistics(cubes, statistics, span):
         result_cube = _compute_eager(aligned_cubes,
                                      operator=operator,
                                      **kwargs)
-
         statistics_cubes[statistic] = result_cube
 
     return statistics_cubes
@@ -328,6 +428,7 @@ def multi_model_statistics(products,
                            span,
                            statistics,
                            output_products=None,
+                           groupby=None,
                            keep_input_datasets=True):
     """Compute multi-model statistics.
 
@@ -357,18 +458,21 @@ def multi_model_statistics(products,
     ----------
     products: list
         Cubes (or products) over which the statistics will be computed.
-    statistics: list
-        Statistical metrics to be computed, e.g. [``mean``, ``max``]. Choose
-        from the operators listed in the iris.analysis package. Percentiles can
-        be specified like ``pXX.YY``.
     span: str
         Overlap or full; if overlap, statitstics are computed on common time-
         span; if full, statistics are computed on full time spans, ignoring
         missing data.
+    statistics: list
+        Statistical metrics to be computed, e.g. [``mean``, ``max``]. Choose
+        from the operators listed in the iris.analysis package. Percentiles can
+        be specified like ``pXX.YY``.
     output_products: dict
         For internal use only. A dict with statistics names as keys and
         preprocessorfiles as values. If products are passed as input, the
         statistics cubes will be assigned to these output products.
+    groupby:  tuple
+        Group products by a given tag or attribute, e.g.
+        ('project', 'dataset', 'tag1').
     keep_input_datasets: bool
         If True, the output will include the input datasets.
         If False, only the computed statistics will be returned.
@@ -393,14 +497,70 @@ def multi_model_statistics(products,
         )
     if all(type(p).__name__ == 'PreprocessorFile' for p in products):
         # Avoid circular input: https://stackoverflow.com/q/16964467
-        return _multiproduct_statistics(
-            products=products,
-            statistics=statistics,
-            output_products=output_products,
-            span=span,
-            keep_input_datasets=keep_input_datasets,
-        )
+        statistics_products = set()
+        for group, input_prods in _group_products(products, by_key=groupby):
+            sub_output_products = output_products[group]
+
+            # Compute statistics on a single group
+            group_statistics = _multiproduct_statistics(
+                products=input_prods,
+                statistics=statistics,
+                output_products=sub_output_products,
+                span=span,
+                keep_input_datasets=keep_input_datasets
+            )
+
+            statistics_products |= group_statistics
+
+        return statistics_products
     raise ValueError(
         "Input type for multi_model_statistics not understood. Expected "
         "iris.cube.Cube or esmvalcore.preprocessor.PreprocessorFile, "
         "got {}".format(products))
+
+
+def ensemble_statistics(products, statistics,
+                        output_products, span='overlap'):
+    """Entry point for ensemble statistics.
+
+    An ensemble grouping is performed on the input products.
+    The statistics are then computed calling
+    the :func:`esmvalcore.preprocessor.multi_model_statistics` module,
+    taking the grouped products as an input.
+
+    Parameters
+    ----------
+    products: list
+        Cubes (or products) over which the statistics will be computed.
+    statistics: list
+        Statistical metrics to be computed, e.g. [``mean``, ``max``]. Choose
+        from the operators listed in the iris.analysis package. Percentiles can
+        be specified like ``pXX.YY``.
+    output_products: dict
+        For internal use only. A dict with statistics names as keys and
+        preprocessorfiles as values. If products are passed as input, the
+        statistics cubes will be assigned to these output products.
+    span: str (default: 'overlap')
+        Overlap or full; if overlap, statitstics are computed on common time-
+        span; if full, statistics are computed on full time spans, ignoring
+        missing data.
+
+    Returns
+    -------
+    set
+        A set of output_products with the resulting ensemble statistics.
+
+    See Also
+    --------
+    :func:`esmvalcore.preprocessor.multi_model_statistics` for
+    the full description of the core statistics function.
+    """
+    ensemble_grouping = ('project', 'dataset', 'exp', 'sub_experiment')
+    return multi_model_statistics(
+        products=products,
+        span=span,
+        statistics=statistics,
+        output_products=output_products,
+        groupby=ensemble_grouping,
+        keep_input_datasets=False
+    )
