@@ -106,20 +106,15 @@ def test_add_aux_coords_from_cubes(coord_dict, output):
 
 
 ALT_COORD = iris.coords.AuxCoord([0.0], bounds=[[-100.0, 500.0]],
-                                 var_name='alt', long_name='altitude',
                                  standard_name='altitude', units='m')
-ALT_COORD_NB = iris.coords.AuxCoord([0.0], var_name='alt',
-                                    long_name='altitude',
-                                    standard_name='altitude', units='m')
+ALT_COORD_NB = iris.coords.AuxCoord([0.0], standard_name='altitude', units='m')
 ALT_COORD_KM = iris.coords.AuxCoord([0.0], bounds=[[-0.1, 0.5]],
                                     var_name='alt', long_name='altitude',
                                     standard_name='altitude', units='km')
 P_COORD = iris.coords.AuxCoord([101325.0], bounds=[[102532.0, 95460.8]],
-                               var_name='plev', standard_name='air_pressure',
-                               long_name='pressure', units='Pa')
-P_COORD_NB = iris.coords.AuxCoord([101325.0], var_name='plev',
-                                  standard_name='air_pressure',
-                                  long_name='pressure', units='Pa')
+                               standard_name='air_pressure', units='Pa')
+P_COORD_NB = iris.coords.AuxCoord([101325.0], standard_name='air_pressure',
+                                  units='Pa')
 CUBE_ALT = iris.cube.Cube([1.0], var_name='x',
                           aux_coords_and_dims=[(ALT_COORD, 0)])
 CUBE_ALT_NB = iris.cube.Cube([1.0], var_name='x',
