@@ -1,13 +1,16 @@
-"""Unit test for :func:`esmvalcore.preprocessor._multimodel`"""
+"""Unit test for :func:`esmvalcore.preprocessor._multimodel`."""
 
 from datetime import datetime
+from unittest import mock
 
 import cftime
 import dask.array as da
 import iris
+import iris.coord_categorisation
 import numpy as np
 import pytest
 from cf_units import Unit
+from iris.coords import AuxCoord
 from iris.cube import Cube
 
 import esmvalcore.preprocessor._multimodel as mm
@@ -46,7 +49,7 @@ def timecoord(frequency,
     elif frequency == 'monthly':
         dates = [datetime(1850, i, 15, 0, 0, 0) for i in time_points]
     elif frequency == 'yearly':
-        dates = [datetime(1850, 7, i, 0, 0, 0) for i in time_points]
+        dates = [datetime(1850 + i - 1, 7, 1, 0, 0, 0) for i in time_points]
 
     unit = Unit(offset, calendar=calendar)
     points = date2num(dates, unit)
@@ -123,6 +126,33 @@ def get_cubes_for_validation_test(frequency, lazy=False):
     return [cube1, cube2, cube3]
 
 
+def get_cube_for_equal_coords_test(num_cubes):
+    """Setup cubes with equal auxiliary coordinates."""
+    cubes = []
+
+    for num in range(num_cubes):
+        cube = generate_cube_from_dates('monthly')
+        cubes.append(cube)
+
+    # Create cubes that have one equal coordinate ('year') and one non-equal
+    # coordinate ('x')
+    year_coord = AuxCoord([1, 2, 3], var_name='year', long_name='year',
+                          units='1', attributes={'test': 1})
+    x_coord = AuxCoord([1, 2, 3], var_name='x', long_name='x', units='s',
+                       attributes={'test': 2})
+    for (idx, cube) in enumerate(cubes):
+        new_x_coord = x_coord.copy()
+        new_x_coord.long_name = f'x_{idx}'
+        cube.add_aux_coord(year_coord.copy(), 0)
+        cube.add_aux_coord(new_x_coord, 0)
+        assert cube.coord('year').metadata is not year_coord.metadata
+        assert cube.coord('year').metadata == year_coord.metadata
+        assert cube.coord(f'x_{idx}').metadata is not x_coord.metadata
+        assert cube.coord(f'x_{idx}').metadata != x_coord.metadata
+
+    return cubes
+
+
 VALIDATION_DATA_SUCCESS = (
     ('full', 'mean', (5, 5, 3)),
     ('full', 'std_dev', (5.656854249492381, 4, 2.8284271247461903)),
@@ -166,6 +196,63 @@ def test_compute_slices(length, slices):
     assert result == slices
 
 
+def test_compute_slices_exceed_end_index():
+    """Test that ``_compute_slices`` terminates when exceeding end index."""
+    # The following settings will result in a cube length of 71, 10 slices and
+    # a slice length of 8. Thus, without early termination, the last slice
+    # would be slice(72, 71), which would result in an exception.
+    cube_data = mock.Mock(nbytes=1.1 * 2**30)  # roughly 1.1 GiB
+    cube = mock.Mock(spec=Cube, data=cube_data, shape=(71,))
+    cubes = [cube] * 9
+
+    slices = list(mm._compute_slices(cubes))
+
+    # Early termination lead to 9 (instead of 10) slices
+    assert len(slices) == 9
+    expected_slices = [
+        slice(0, 8, None),
+        slice(8, 16, None),
+        slice(16, 24, None),
+        slice(24, 32, None),
+        slice(32, 40, None),
+        slice(40, 48, None),
+        slice(48, 56, None),
+        slice(56, 64, None),
+        slice(64, 71, None),
+    ]
+    assert slices == expected_slices
+
+
+def test_compute_slices_equals_end_index():
+    """Test that ``_compute_slices`` terminates when reaching end index."""
+    # The following settings will result in a cube length of 36, 13 slices and
+    # a slice length of 3. Thus, without early termination, the last slice
+    # would be slice(36, 39), which would result in an exception.
+    cube_data = mock.Mock(nbytes=1.05 * 2**30)  # roughly 1.05 GiB
+    cube = mock.Mock(spec=Cube, data=cube_data, shape=(36,))
+    cubes = [cube] * 12
+
+    slices = list(mm._compute_slices(cubes))
+
+    # Early termination lead to 12 (instead of 13) slices
+    assert len(slices) == 12
+    expected_slices = [
+        slice(0, 3, None),
+        slice(3, 6, None),
+        slice(6, 9, None),
+        slice(9, 12, None),
+        slice(12, 15, None),
+        slice(15, 18, None),
+        slice(18, 21, None),
+        slice(21, 24, None),
+        slice(24, 27, None),
+        slice(27, 30, None),
+        slice(30, 33, None),
+        slice(33, 36, None),
+    ]
+    assert slices == expected_slices
+
+
 @pytest.mark.parametrize('frequency', FREQUENCY_OPTIONS)
 @pytest.mark.parametrize('span, statistics, expected', VALIDATION_DATA_SUCCESS)
 def test_multimodel_statistics(frequency, span, statistics, expected):
@@ -371,6 +458,56 @@ def test_combine_inconsistent_var_names_fail():
         _ = mm._combine(cubes)
 
 
+@pytest.mark.parametrize('scalar_coord', ['p0', 'ptop'])
+def test_combine_with_scalar_coords_to_remove(scalar_coord):
+    """Test _combine with scalar coordinates that should be removed."""
+    num_cubes = 5
+    cubes = []
+
+    for num in range(num_cubes):
+        cube = generate_cube_from_dates('monthly')
+        cubes.append(cube)
+
+    scalar_coord_0 = AuxCoord(0.0, var_name=scalar_coord)
+    scalar_coord_1 = AuxCoord(1.0, var_name=scalar_coord)
+    cubes[0].add_aux_coord(scalar_coord_0, ())
+    cubes[1].add_aux_coord(scalar_coord_1, ())
+
+    merged_cube = mm._combine(cubes)
+    assert merged_cube.shape == (5, 3)
+
+
+def test_combine_preserve_equal_coordinates():
+    """Test ``_combine`` with equal input coordinates."""
+    cubes = get_cube_for_equal_coords_test(5)
+    merged_cube = mm._combine(cubes)
+
+    # The equal coordinate ('year') was not changed; the non-equal one ('x')
+    # does not have a long_name and attributes anymore
+    assert merged_cube.coord('year').var_name == 'year'
+    assert merged_cube.coord('year').standard_name is None
+    assert merged_cube.coord('year').long_name == 'year'
+    assert merged_cube.coord('year').attributes == {'test': 1}
+    assert merged_cube.coord('x').var_name == 'x'
+    assert merged_cube.coord('x').standard_name is None
+    assert merged_cube.coord('x').long_name is None
+    assert merged_cube.coord('x').attributes == {}
+
+
+def test_equalise_coordinates_no_cubes():
+    """Test that _equalise_coordinates doesn't fail with empty cubes."""
+    mm._equalise_coordinates([])
+
+
+def test_equalise_coordinates_one_cube():
+    """Test that _equalise_coordinates doesn't fail with a single cubes."""
+    cube = generate_cube_from_dates('monthly')
+    new_cube = cube.copy()
+    mm._equalise_coordinates([new_cube])
+    assert new_cube is not cube
+    assert new_cube == cube
+
+
 @pytest.mark.parametrize('span', SPAN_OPTIONS)
 def test_edge_case_different_time_offsets(span):
     cubes = (
@@ -514,13 +651,37 @@ def test_unify_time_coordinates():
 class PreprocessorFile:
     """Mockup to test output of multimodel."""
 
-    def __init__(self, cube=None):
+    def __init__(self, cube=None, attributes=None):
         if cube:
             self.cubes = [cube]
+        if attributes:
+            self.attributes = attributes
 
     def wasderivedfrom(self, product):
         pass
 
+    def group(self, keys: list) -> str:
+        """Generate group keyword.
+
+        Returns a string that identifies a group. Concatenates a list of
+        values from .attributes
+        """
+        if not keys:
+            return ''
+
+        if isinstance(keys, str):
+            keys = [keys]
+
+        identifier = []
+        for key in keys:
+            attribute = self.attributes.get(key)
+            if attribute:
+                if isinstance(attribute, (list, tuple)):
+                    attribute = '-'.join(attribute)
+                identifier.append(attribute)
+
+        return '_'.join(identifier)
+
 
 def test_return_products():
     """Check that the right product set is returned."""
@@ -533,12 +694,12 @@ def test_return_products():
     products = set([input1, input2])
 
     output = PreprocessorFile()
-    output_products = {'mean': output}
+    output_products = {'': {'mean': output}}
 
     kwargs = {
         'statistics': ['mean'],
         'span': 'full',
-        'output_products': output_products
+        'output_products': output_products['']
     }
 
     result1 = mm._multiproduct_statistics(products,
@@ -552,6 +713,7 @@ def test_return_products():
     assert result1 == set([input1, input2, output])
     assert result2 == set([output])
 
+    kwargs['output_products'] = output_products
     result3 = mm.multi_model_statistics(products, **kwargs)
     result4 = mm.multi_model_statistics(products,
                                         keep_input_datasets=False,
@@ -561,6 +723,48 @@ def test_return_products():
     assert result4 == result2
 
 
+def test_ensemble_products():
+    cube1 = generate_cube_from_dates('monthly', fill_val=1)
+    cube2 = generate_cube_from_dates('monthly', fill_val=9)
+
+    attributes1 = {
+        'project': 'project', 'dataset': 'dataset',
+        'exp': 'exp', 'ensemble': '1'}
+    input1 = PreprocessorFile(cube1, attributes=attributes1)
+
+    attributes2 = {
+        'project': 'project', 'dataset': 'dataset',
+        'exp': 'exp', 'ensemble': '2'}
+    input2 = PreprocessorFile(cube2, attributes=attributes2)
+
+    attributes3 = {
+        'project': 'project', 'dataset': 'dataset2',
+        'exp': 'exp', 'ensemble': '1'}
+    input3 = PreprocessorFile(cube1, attributes=attributes3)
+
+    attributes4 = {
+        'project': 'project', 'dataset': 'dataset2',
+        'exp': 'exp', 'ensemble': '2'}
+
+    input4 = PreprocessorFile(cube1, attributes=attributes4)
+    products = set([input1, input2, input3, input4])
+
+    output1 = PreprocessorFile()
+    output2 = PreprocessorFile()
+    output_products = {
+        'project_dataset_exp': {'mean': output1},
+        'project_dataset2_exp': {'mean': output2}}
+
+    kwargs = {
+        'statistics': ['mean'],
+        'output_products': output_products,
+    }
+
+    result = mm.ensemble_statistics(
+        products, **kwargs)
+    assert len(result) == 2
+
+
 def test_ignore_tas_scalar_height_coord():
     """Ignore conflicting aux_coords for height in tas."""
     tas_2m = generate_cube_from_dates("monthly")
@@ -670,3 +874,50 @@ def test_no_warn_model_dim_non_contiguous(recwarn):
            "Metadata may not be fully descriptive for 'multi-model'.")
     for warning in recwarn:
         assert str(warning.message) != msg
+
+
+def test_map_to_new_time_int_coords():
+    """Test ``_map_to_new_time`` with integer time coords."""
+    cube = generate_cube_from_dates('yearly')
+    iris.coord_categorisation.add_year(cube, 'time')
+    decade_coord = AuxCoord([1850, 1850, 1850], bounds=[[1845, 1855]] * 3,
+                            long_name='decade')
+    cube.add_aux_coord(decade_coord, 0)
+    target_points = [200.0, 500.0, 1000.0]
+
+    out_cube = mm._map_to_new_time(cube, target_points)
+
+    assert_array_allclose(out_cube.data,
+                          np.ma.masked_invalid([1.0, 1.0, np.nan]))
+    assert_array_allclose(out_cube.coord('time').points, target_points)
+    assert_array_allclose(out_cube.coord('year').points,
+                          np.ma.masked_invalid([1850, 1851, np.nan]))
+    assert_array_allclose(out_cube.coord('decade').points,
+                          np.ma.masked_invalid([1850, 1850, np.nan]))
+    assert out_cube.coord('year').bounds is None
+    assert out_cube.coord('decade').bounds is None
+    assert np.issubdtype(out_cube.coord('year').dtype, np.integer)
+    assert np.issubdtype(out_cube.coord('decade').dtype, np.integer)
+
+
+def test_preserve_equal_coordinates():
+    """Test ``multi_model_statistics`` with equal input coordinates."""
+    cubes = get_cube_for_equal_coords_test(5)
+    stat_cubes = multi_model_statistics(cubes, span='overlap',
+                                        statistics=['sum'])
+
+    assert len(stat_cubes) == 1
+    assert 'sum' in stat_cubes
+    stat_cube = stat_cubes['sum']
+    assert_array_allclose(stat_cube.data, np.ma.array([5.0, 5.0, 5.0]))
+
+    # The equal coordinate ('year') was not changed; the non-equal one ('x')
+    # does not have a long_name and attributes anymore
+    assert stat_cube.coord('year').var_name == 'year'
+    assert stat_cube.coord('year').standard_name is None
+    assert stat_cube.coord('year').long_name == 'year'
+    assert stat_cube.coord('year').attributes == {'test': 1}
+    assert stat_cube.coord('x').var_name == 'x'
+    assert stat_cube.coord('x').standard_name is None
+    assert stat_cube.coord('x').long_name is None
+    assert stat_cube.coord('x').attributes == {}
